import copy
import re
import sys

import boto3
import botocore.exceptions

import skwadon.main as sic_main
import skwadon.common_action as common_action
import skwadon.aws_iam_group        as aws_iam_group
import skwadon.aws_iam_role         as aws_iam_role
import skwadon.aws_iam_policy       as aws_iam_policy
import skwadon.aws_s3_bucket        as aws_s3_bucket
import skwadon.aws_stepfunctions    as aws_stepfunctions
import skwadon.aws_lambda           as aws_lambda
import skwadon.aws_glue_datacatalog as aws_glue_datacatalog
import skwadon.aws_glue_crawler     as aws_glue_crawler
import skwadon.aws_glue_job         as aws_glue_job
import skwadon.aws_glue_connection  as aws_glue_connection
import skwadon.aws_glue_trigger     as aws_glue_trigger
import skwadon.aws_rds              as aws_rds
import skwadon.aws_redshift         as aws_redshift

def get_message_prefix(data):
    if "profile" in data:
        profile = data["profile"]
    else:
        profile = "default"
    ret = f"aws(proifle={profile}"
    if "region" in data:
        region = data["region"]
        ret = ret + f", region={region}"
    ret = ret + ")"
    return ret

def get_handler(src_data):
    session = create_aws_session(src_data)
    return common_action.NamespaceHandler(
        None, [], {
        "iam.groups": aws_iam_group.GroupListHandler(session),
        "iam.roles": aws_iam_role.RoleListHandler(session),
        "iam.policies": aws_iam_policy.PolicyListHandler(session),
        "s3.buckets": aws_s3_bucket.BucketListHandler(session),
        "lambda.functions": aws_lambda.FunctionListHandler(session),
        "stepFunctions.stateMachines": aws_stepfunctions.StateMachineListHandler(session),
        "glue.crawlers": aws_glue_crawler.CrawlerListHandler(session),
        "glue.databases": aws_glue_datacatalog.DatabaseListHandler(session),
        "glue.jobs": aws_glue_job.JobListHandler(session),
        "glue.connections": aws_glue_connection.ConnectionListHandler(session),
        "glue.triggers": aws_glue_trigger.TriggerListHandler(session),
        "rds.instances": aws_rds.InstanceListHandler(session),
        "redshift.clusters": aws_redshift.ClusterListHandler(session),
    })

def create_aws_session(data):
    if "profile" in data:
        profile = data["profile"]
    else:
        profile = "default"
    if "region" in data:
        region = data["region"]
    else:
        region = None
    session = boto3.session.Session(profile_name = profile, region_name = region)
    return session

def fetch_account_id(session):
    account_id = session.client("sts").get_caller_identity()["Account"]
    return account_id

def fetch_region_name(session):
    region_name = session.region_name
    if region_name == None:
        raise Exception()
    return region_name

def fetch_s3_object(session, s3_path: str):
    if s3_path == None:
        return None
    s3_client = session.client("s3")
    m = re.compile("\As3://([^/]+)/(.*)\Z").search(s3_path)
    if not m:
        return None
    s3_bucket = m.group(1)
    s3_key = m.group(2)
    try:
        res = s3_client.get_object(Bucket = s3_bucket, Key = s3_key)
        body = res['Body'].read()
        body_str = body.decode('utf-8')
        return body_str
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchKey":
            return None
        else:
            raise

def put_s3_object(confirmation_flag, session, s3_path: str, body: str):
    if s3_path == None:
        return
    s3_client = session.client("s3")
    m = re.compile("\As3://([^/]+)/(.*)\Z").search(s3_path)
    if not m:
        return
    s3_bucket = m.group(1)
    s3_key = m.group(2)
    sic_main.exec_put(confirmation_flag,
        f"s3_client.put_object(Bucket = {s3_bucket}, Key = {s3_key}, ...)",
        lambda:
            s3_client.put_object(Bucket = s3_bucket, Key = s3_key, Body = body.encode('utf-8'))
    )



import copy
import datetime
import sys
import yaml

import skwadon.lib as sic_lib
import skwadon.testcloud as testcloud
import skwadon.aws as sic_aws

####################################################################################################

def main():
    # interprete parameters
    (help_flag, action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm) = parse_args()

    if help_flag:
        print_help()
        return

    # check parameters
    (action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm) = check_args \
        (action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm)

    # execution
    exec_main \
        (action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm)

####################################################################################################
# interprete parameters
# パラメータ解釈
####################################################################################################

def parse_args():
    help_flag = False
    action = None # get, put, delete exec
    is_simple = None
    is_full = None
    is_diff = None
    is_completion = None
    thats_all_flag = False
    type = None # testcloud, aws
    profile = None
    path = []
    src_file = None
    is_inplace = False
    is_dryrun = False
    confirm = None
    i = 1
    argCount = len(sys.argv)
    while i < argCount:
        a = sys.argv[i]
        i = i + 1
        if a == "--help":
            # ヘルプはまだ実装していない
            help_flag = True
        elif a == "-r":
            is_simple = True
        elif a == "--full":
            is_full = True
        elif a == "--diff":
            is_diff = True
        elif a == "--thats-all":
            thats_all_flag = True
        elif a == "--no-diff":
            is_diff = False
        elif a == "--profile":
            if i >= argCount:
                raise Exception(f"Option parameter not found: {a}")
            profile = sys.argv[i]
            i = i + 1
        elif a == "-p":
            if i >= argCount:
                raise Exception(f"Option parameter not found: {a}")
            path.append(sys.argv[i])
            i = i + 1
        elif a == "-s":
            if i >= argCount:
                raise Exception(f"Option parameter not found: {a}")
            src_file = sys.argv[i]
            i = i + 1
        elif a == "-i":
            is_inplace = True
        elif a == "--force":
            confirm = True
        elif a == "--dry-run":
            is_dryrun = True
        elif a == "--confirm":
            if i >= argCount:
                # エラーだけど
                # check_confirm で親切なエラーを出すためここではエラー扱いにしない
                confirm = ""
            else:
                confirm = sys.argv[i]
                i = i + 1
        elif a.startswith("-"):
            raise Exception(f"Unknown option: {a}")

        # actionはこの4種類
        elif action is None and a == "get":
            action = "get"
        elif action is None and a == "put":
            action = "put"
        elif action is None and a == "delete":
            action = "delete"
        elif action is None and a == "exec":
            action = "exec"

        # クラウド種類はいまのところこの2種類
        elif a == "testcloud":
            # skwadonテストコードのためのクラウド種類
            type = "testcloud"
        elif a == "aws":
            # AWS
            type = "aws"

        elif type and src_file is None:
            path.append(a)
        elif type is None and path is None and src_file is None:
            src_file = a

        else:
            raise Exception(f"Unknown parameter: {a}")

    return (help_flag, action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm)

####################################################################################################
# check parameters
# パラメータの組み合わせチェック
####################################################################################################

def check_args(action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm):
    if len(path) > 0 and type is None:
        raise Exception("-p option needs aws parameter")

    # pathが . で終わっている場合はその次に続く文字列候補を出力する
    if len(path) > 0 and path[len(path) - 1].endswith("."):
        last_path = path[len(path) - 1]
        path = [last_path[0:len(last_path) - 1]]
        action = "get"
        is_simple = False
        is_full = False
        is_diff = False
        is_completion = True
        thats_all_flag = False
        src_file = None
        is_inplace = False

    # pathが複数ある場合は --full とする
    if len(path) > 1:
        is_simple = False
        is_full = True

    if action == "put" and len(path) > 1:
        raise Exception("put action must have only one path")

    if action == "exec" and len(path) > 1:
        raise Exception("exec action must have only one path")
    if action == "exec" and len(path) == 0:
        raise Exception("exec action must have one path")

    # actionの指定がない場合は get とみなす
    if action is None:
        action = "get"
        if type is None and src_file is None:
            action = "cat"

    # is_diffの指定がない場合のデフォルト値設定
    if action == "get" or action == "cat":
        if is_diff is None:
            is_diff = False
    elif action == "put" or action == "delete":
        if is_diff and not is_dryrun and not confirm:
            # --diff が指定されていて --dry-run も --confirm もない場合は --dry-run とみなす
            is_dryrun = True
        if is_dryrun:
            if is_diff is None:
                is_diff = True
        else:
            # put実行では差分表示もフル表示もしない
            pass

    if action == "delete":
        if type is None:
            raise Exception(f"delete action needs aws parameter")
        if src_file is not None:
            raise Exception(f"delete action must not have -s option")
        if len(path) == 0:
            raise Exception(f"delete action needs -p option")

    # 入力がない場合はエラー
    if type is None and src_file is None and sys.stdin.isatty():
        raise Exception(f"either aws parameter or -s must be expected")

    # ファイルでの入力の場合の出力は is_full のみ
    if type is None:
        is_simple = False
        is_full = True

    if is_inplace:
        if src_file is None:
            raise Exception("-i option needs -s option")
        if action != "get" and action != "cat":
            raise Exception(f"{action} must not have -i option")
        if is_diff:
            raise Exception(f"only one of --diff and -i can be specified")

    if type is not None:
        path2 = []
        for p in path:
            if p == "":
                path2.append([])
            else:
                if p.endswith("."):
                    p = p[0:len(p)-1]
                path2.append(p.split("."))
        if len(path) == 0 and src_file is None:
            path2.append([])
        path = path2

    return (action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm)

####################################################################################################
# ヘルプ
####################################################################################################

def print_help():
    help_str = """
skwadon [OPTIONS]

OPTION:
    --help
    --profile <AWS_PROFILE_NAME>
    --region <AWS_REGION_NAME>
    ...
""".strip()
    print(help_str)

####################################################################################################
# execution
# 実行
####################################################################################################

def exec_main(action, is_simple, is_full, is_diff, is_completion, thats_all_flag, type, profile, path, src_file, is_dryrun, is_inplace, confirm):
    global global_confirmation_flag
    confirmation_flag = False

    if action == "put" or action == "delete":
        if not is_dryrun:
            # putコマンドでは --confirm オプションをチェック
            if confirm is None:
                hm = get_correct_confirm_parameter()
                raise Exception(f"put action needs --dry-run or --confirm {hm}")
            if confirm != True:
                check_confirm(confirm)
            confirmation_flag = True
            global_confirmation_flag = True

    if type is not None:
        # get aws -p ... < data.yml
        # put aws -p ... < data.yml
        # のパターン
        if action == "delete":
            data_put = None # 削除の意味
        elif action == "exec":
            data_put = None
        elif action == "put":
            data_put = load_simple(src_file, {})
            # 標準入力がない場合は {} になる
        else:
            data_put = load_simple(src_file, None)
            # 標準入力がない場合は null になる
        data0 = build_path_data_full(type, profile, path, data_put)
    else:
        data0 = load_yaml(src_file)

    if action == "delete":
        action = "put"

    if action == "cat":
        data1 = data0
        r1 = data0
        r2 = data1
        action = "get"
    elif action == "get":
        data1 = do_get_n(data0, thats_all_flag)
        r1 = data0 # src
        r2 = data1 # クラウド側
    elif action == "put":
        data0, data1 = do_put_n(confirmation_flag, data0)
        r1 = data1 # 更新前のクラウド側
        r2 = data0 # srcまたは更新後
        if confirmation_flag:
            add_update_completion_message()
    elif action == "exec":
        do_exec(data0)

    if action == "exec":
        pass # ここでは出力制御しない
    elif is_completion:
        output_completion(get_by_path(data1, path))
    elif action == "get":
        if is_full:
            if is_diff:
                diff_yaml(r1, r2)
            elif is_inplace:
                save_yaml(data1, src_file)
            else:
                save_yaml(data1, None)
        elif is_simple:
            if is_diff:
                output_simple_diff(get_by_path(r1, path), get_by_path(r2, path))
            else:
                output_simple(get_by_path(data1, path))
        else:
            if is_diff:
                diff_yaml(get_by_path(r1, path), get_by_path(r2, path))
            else:
                save_yaml(get_by_path(data1, path), None)
    elif action == "put":
        if is_full:
            if is_diff:
                diff_yaml(r1, r2)
            else:
                pass
        elif is_simple:
            if is_diff:
                output_simple_diff(get_by_path(r1, path), get_by_path(r2, path))
            else:
                pass
        else:
            if is_diff:
                diff_yaml(get_by_path(r1, path), get_by_path(r2, path))
            else:
                pass

####################################################################################################

# --confirm の時間チェック
def check_confirm(confirm):
    now = datetime.datetime.now(datetime.timezone.utc)
    for i in range(3):
        time_str = (now + datetime.timedelta(minutes = i - 1)).isoformat()
        hm = time_str[11:13] + time_str[14:16]
        if hm == confirm:
            return True
    hm = get_correct_confirm_parameter()
    raise Exception(f"put action needs --confirm {hm}")

def get_correct_confirm_parameter():
    now = datetime.datetime.now(datetime.timezone.utc)
    time_str = now.isoformat()
    return time_str[11:13] + time_str[14:16]

####################################################################################################

# -p オプションからデータ作成
def build_path_data_full(type, profile, path, data_put):
    data0 = {}
    data0["type"] = type
    if profile:
        data0["profile"] = profile
    data0["resources"] = build_path_data(path, data_put)
    return data0

def build_path_data(path, data_put):
    def merge(d1, d2):
        if not isinstance(d1, dict) or not isinstance(d2, dict):
            return d2
        result = {}
        for k in d1:
            if k in d2:
                result[k] = merge(d1[k], d2[k])
            else:
                result[k] = d1[k]
        for k in d2:
            if not k in d1:
                result[k] = d2[k]
        return result
    data = {}
    for p in path:
        d = build_path_data_one(p, copy.copy(data_put))
        data = merge(data, d)
    return data

def build_path_data_one(path, data_put):
    data = data_put
    for elem in reversed(path):
        data1 = {}
        data1[elem] = data
        data = data1
    return data

# データから -p で指定された場所を抜き出す
def get_by_path(data, path):
    if len(path) != 1:
        return data
    path = path[0]
    if "*" in path:
        return data
    def sub(data):
        if path is None:
            result = data
        else:
            result = data["resources"]
            for elem in path:
                if result is None:
                    return None
                if not elem in result:
                    return None
                result = result[elem]
        return result
    if isinstance(data, list):
        result = []
        for elem in data:
            result.append(sub(elem))
    else:
        result = sub(data)
    return result

def load_yaml(src_file):
    if src_file:
        with open(src_file) as f:
            data = yaml.safe_load(f)
    elif sys.stdin.isatty():
        #raise Exception(f"-s not specified")
        data = {}
    else:
        data = yaml.safe_load(sys.stdin)
    return data

def load_simple(src_file, default):
    if src_file:
        with open(src_file) as f:
            data_str = f.read()
    elif sys.stdin.isatty():
        return default
    else:
        data_str = sys.stdin.read()
    data = yaml.safe_load(data_str)
    return data

####################################################################################################

def output_completion(result):
    if result is None:
        pass
    elif isinstance(result, dict):
        max_len = 0
        names = []
        for name, value in result.items():
            if name == "*":
                continue
            if len(name) > max_len:
                max_len = len(name)
            names.append(name)
        for name, msg in help_message.items():
            if name in names:
                continue
            if len(name) > max_len:
                max_len = len(name)
            names.append(name)
        for name in names:
            if name in help_message:
                msg = help_message[name]
            else:
                msg = ""
            if len(msg) > 0:
                padding = " " * (max_len - len(name))
                print(f"{name}{padding} : {msg}")
            else:
                print(f"{name}")
    elif isinstance(result, list):
        is_str_list = True
        for elem in result:
            if not isinstance(elem, str):
                is_str_list = False
                break
        if is_str_list:
            for name in result:
                print(name)

def output_simple(result):
    if result is None:
        pass
    elif isinstance(result, dict):
        is_simple = True
        for name, value in result.items():
            if value != {}:
                is_simple = False
        if is_simple:
            for name, value in result.items():
                print(name)
        else:
            save_yaml(result, None)
    elif isinstance(result, list):
        is_str_list = True
        for elem in result:
            if not isinstance(elem, str):
                is_str_list = False
                break
        if is_str_list:
            for name in result:
                print(name)
        else:
            save_yaml(result, None)
    elif isinstance(result, str):
        print(result)
    elif isinstance(result, int):
        print(result)
    else:
        save_yaml(result, None)

def output_simple_diff(result1, result2):
    if result1 == {} and (isinstance(result2, str) or isinstance(result2, int)):
        result1 = ""
    if (result1 == "" or isinstance(result1, str)) and isinstance(result2, str):
        sic_lib.exec_diff(result1, result2, None)
    elif (result1 == "" or isinstance(result1, int)) and isinstance(result2, int):
        sic_lib.exec_diff(str(result1), str(result2), None)
    else:
        diff_yaml(result1, result2)

def represent_str(dumper, s):
    if "\n" in s:
        return dumper.represent_scalar('tag:yaml.org,2002:str', s, style='|')
    else:
        return dumper.represent_scalar('tag:yaml.org,2002:str', s)

yaml.add_representer(str, represent_str)

def save_yaml(data, dst_file):
    yaml_str = yaml.dump(data, sort_keys = False, allow_unicode = True, width = 120, default_flow_style = False)
    if dst_file:
        with open(dst_file, "w") as f:
            f.write(yaml_str)
    else:
        sys.stdout.write(yaml_str)

def diff_yaml(src_data, dst_data):
    src_yaml_str = yaml.dump(src_data, sort_keys = False, allow_unicode = True, width = 120)
    dst_yaml_str = yaml.dump(dst_data, sort_keys = False, allow_unicode = True, width = 120)
    sic_lib.exec_diff(src_yaml_str, dst_yaml_str, None)

####################################################################################################

def do_get_n(data0, thats_all_flag):
    data1 = data0
    if isinstance(data1, list):
        data2 = []
        for elem in data1:
            r = do_get(elem, thats_all_flag)
            data2.append(r)
    else:
        data2 = do_get(data1, thats_all_flag)
    return data2

def do_get(src_data, thats_all_flag):
    global update_message_prefix
    result = src_data
    if src_data["type"] == "testcloud":
        handler = testcloud.get_handler(src_data)
        result = copy.copy(src_data)
        if "resources" in src_data:
            result["resources"] = handler.do_get(src_data["resources"], thats_all_flag)
    elif src_data["type"] == "aws":
        handler = sic_aws.get_handler(src_data)
        result = copy.copy(src_data)
        if "resources" in src_data:
            result["resources"] = handler.do_get(src_data["resources"], thats_all_flag)
    return result

def do_put_n(confirmation_flag, data0):
    if isinstance(data0, list):
        data1 = []
        data2 = []
        for elem in data0:
            d1, d2 = do_put(confirmation_flag, elem)
            data1.append(d1)
            data2.append(d2)
    else:
        data1, data2 = do_put(confirmation_flag, data0)
    return (data1, data2)

def do_put(confirmation_flag, src_data):
    global update_message_prefix
    mod = None
    if src_data["type"] == "testcloud":
        mod = testcloud
    elif src_data["type"] == "aws":
        mod = sic_aws
    if mod:
        update_message_prefix = mod.get_message_prefix(src_data)
        handler = mod.get_handler(src_data)
        data1 = copy.copy(src_data)
        data2 = copy.copy(src_data)
        if "resources" in src_data:
            d1, d2 = handler.do_put(confirmation_flag, src_data["resources"])
            data1["resources"] = d1
            data2["resources"] = d2
        update_message_prefix = None
        return (data1, data2)
    else:
        return (src_data, src_data)

def do_exec(src_data):
    mod = None
    if src_data["type"] == "testcloud":
        mod = testcloud
    elif src_data["type"] == "aws":
        mod = sic_aws
    if mod:
        handler = mod.get_handler(src_data)
        if "resources" in src_data:
            raise Exception("TODO")
            #handler.do_exec(src_data["resources"])

####################################################################################################

# boto3で変更のAPIリクエストを実行する際には必ずこのメソッドを通すこと
# dryrunでも通す
# dryrunかどうかはconfirmation_flagで判定
def exec_put(confirmation_flag, message, executor):
    add_update_message(message)
    if confirmation_flag and global_confirmation_flag:
        executor()

# putアクションで --dry-run が指定されていなく、更新処理を実行することを示すフラグ
# バグにより意図せず更新処理してしまうのが怖いので、引き回しせずにグローバルで持つことにする
global_confirmation_flag = False

update_message_prefix = None
update_message = []

# 更新系APIコールの直前で呼ばれる
# --dry-run でも呼ばれる
def add_update_message(message):
    if update_message_prefix:
        message = f"{update_message_prefix}: {message}"
    else:
        message = f"{message}"
    #print(message, file = sys.stderr)
    print(message)
    update_message.append(message)

# 更新系の処理がすべて完了したら呼ばれる
# --dry-run では呼ばれない
def add_update_completion_message():
    if len(update_message) > 0:
        message = "complete put action"
        #print(message, file = sys.stderr)
        print(message)

help_message = {}

def put_help_message(key, message):
    help_message[key] = message

####################################################################################################



import copy
import json
import re

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action

class ConnectionListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.glue_client = None

    def init_client(self):
        if self.glue_client == None:
            self.glue_client = self.session.client("glue")

    def list(self):
        self.init_client()
        result = []
        res = self.glue_client.get_connections()
        while True:
            for elem in res['ConnectionList']:
                name = elem["Name"]
                result.append(name)
            if not "NextToken" in res:
                break
            res = self.glue_client.get_connections(NextToken = res["NextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf"], {
            "conf": ConnectionConfHandler(self.glue_client, name),
            "connection": ConnectionConnectionHandler(self.glue_client, name),
        })

class ConnectionConfHandler(common_action.ResourceHandler):

    properties = [
        "Description",
        "ConnectionType",
        "MatchCriteria",
        "ConnectionProperties",
        "PhysicalConnectionRequirements",
    ]

    def __init__(self, glue_client, connection_name):
        self.glue_client = glue_client
        self.connection_name = connection_name

    def describe(self):
        try:
            res = self.glue_client.get_connection(Name = self.connection_name)
            curr_data = sic_lib.pickup(res["Connection"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = sic_lib.pickup(src_data, self.properties)
        update_data["Name"] = self.connection_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_connection(ConnectionInput = {{Name = {self.connection_name}, ...}})",
            lambda:
                self.glue_client.create_connection(ConnectionInput = update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickupAndCompareForUpdate(src_data, curr_data, self.properties)
        if update_data != None:
            update_data["Name"] = self.connection_name
            sic_main.exec_put(confirmation_flag,
                f"glue_client.update_connection(Name = {self.connection_name}, ...)",
                lambda:
                    self.glue_client.update_connection(Name = self.connection_name, ConnectionInput = update_data)
            )

class ConnectionConnectionHandler(common_action.ResourceHandler):

    properties = [
        "Description",
        "ConnectionType",
        "MatchCriteria",
        "ConnectionProperties",
        "PhysicalConnectionRequirements",
    ]

    def __init__(self, glue_client, connection_name):
        self.glue_client = glue_client
        self.connection_name = connection_name

    def describe(self):
        res = self.glue_client.get_connection(Name = self.connection_name)
        jdbc_url = sic_lib.dictlib_get(res, "Connection.ConnectionProperties.JDBC_CONNECTION_URL")
        curr_data = {
        }
        if jdbc_url != None:
            curr_data["JBCDUrl"] = jdbc_url
            m = re.compile("\Ajdbc:([^:]+)://([^:]+):([0-9]+)/(.+)\Z").search(jdbc_url)
            if m:
                driver = m.group(1)
                if driver == "postgresql" or driver == "redshift":
                    host = m.group(2)
                    port = m.group(3)
                    db_name = m.group(4)
                    user_name = sic_lib.dictlib_get(res, "Connection.ConnectionProperties.USERNAME")
                    password = sic_lib.dictlib_get(res, "Connection.ConnectionProperties.PASSWORD")
                    cmd = f"PGPASSWORD='{password}' psql -h {host} -p {port} -U {user_name} -d {db_name}"
                    curr_data["CommandLine"] = cmd
        return curr_data

    def create(self, confirmation_flag, src_data):
        pass

    def update(self, confirmation_flag, src_data, curr_data):
        pass



import json

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as sic_aws

class StateMachineListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.stepfunctions_client = None

    def init_client(self):
        if self.stepfunctions_client == None:
            self.stepfunctions_client = self.session.client("stepfunctions")

    def list(self):
        self.init_client()
        account_id = sic_aws.fetch_account_id(self.session)
        region_name = sic_aws.fetch_region_name(self.session)
        result = []
        res = self.stepfunctions_client.list_state_machines()
        while True:
            for elem in res['stateMachines']:
                name = elem["name"]
                arn = calc_statemachine_arn2(account_id, region_name, name)
                if elem["stateMachineArn"] == arn:
                    result.append(name)
            if not "nextToken" in res:
                break
            res = self.stepfunctions_client.list_state_machines(nextToken = res["nextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "definition"], {
            "conf": StateMachineConfHandler(self.session, self.stepfunctions_client, name),
            "definition": StateMachineDefinitionHandler(self.session, self.stepfunctions_client, name),
        })

class StateMachineBasicHandler(common_action.ResourceHandler):
    def __init__(self, session, stepfunctions_client, machine_name):
        self.session = session
        self.stepfunctions_client = stepfunctions_client
        self.machine_name = machine_name
        self.arn = None
        self.info = None

    def _describe(self):
        if self.arn != None:
            return self.info
        if self.arn == None:
            arn = self._calc_statemachine_arn()
        try:
            res = self.stepfunctions_client.describe_state_machine(stateMachineArn = arn)
            curr_data = sic_lib.dict_key_capitalize(res)
            curr_data = sic_lib.pickup(curr_data, self._properties)
            self.info = curr_data
            return self.info
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "StateMachineDoesNotExist":
                return None
            else:
                raise

    def _create(self, confirmation_flag, src_data):
        update_data = {}
        update_data["Name"] = self.machine_name
        update_data["Definition"] = self._defaultDefinition()
        for key in self._properties_for_create:
            if key in src_data:
                update_data[key] = src_data[key]
        update_data = sic_lib.dict_key_to_lower(update_data)
        sic_main.exec_put(confirmation_flag,
            f"stepfunctions_client.create_state_machine(name = {self.machine_name}, ...)",
            lambda:
                self.stepfunctions_client.create_state_machine(**update_data)
        )
        if confirmation_flag:
            self.info = None

    def _update(self, confirmation_flag, src_data):
        curr_data = self._describe()
        if curr_data is None:
            curr_data = {}
        update_data = {}
        dirty = False
        for key in self._properties_for_update:
            if key in src_data:
                update_data[key] = src_data[key]
                if key not in curr_data or update_data[key] != curr_data[key]:
                    dirty = True
            elif key in curr_data:
                update_data[key] = curr_data[key]
        if dirty:
            arn = self._calc_statemachine_arn()
            update_data = sic_lib.dict_key_to_lower(update_data)
            sic_main.exec_put(confirmation_flag,
                f"stepfunctions_client.update_state_machine(stateMachineArn = {arn}, ...)",
                lambda:
                    self.stepfunctions_client.update_state_machine(stateMachineArn = arn, **update_data)
            )
            if confirmation_flag:
                self.info = None

    def _calc_statemachine_arn(self):
        return calc_statemachine_arn(self.session, self.machine_name)

    def _defaultDefinition(self):
        return json.dumps({
            "StartAt": "__skwadon_dummy",
            "States": {
                "__skwadon_dummy": {
                    "Type": "Pass",
                    "End": True,
                },
            },
        })

    _properties = [
        "RoleArn",
        "Type",
        "LoggingConfiguration",
        "TracingConfiguration",
        "Definition"
    ]
    _properties_for_create = [
        "RoleArn",
        "Type",
        "LoggingConfiguration",
        "TracingConfiguration",
    ]
    _properties_for_update = [
        "RoleArn",
        "LoggingConfiguration",
        "TracingConfiguration",
        "Definition"
    ]


class StateMachineConfHandler(StateMachineBasicHandler):
    def __init__(self, session, stepfunctions_client, machine_name):
        StateMachineBasicHandler.__init__(self, session, stepfunctions_client, machine_name)

    def describe(self):
        info = self._describe()
        if info == None:
            return None
        curr_data = sic_lib.pickup(info, self.properties)
        return curr_data

    def create(self, confirmation_flag, src_data):
        self._create(confirmation_flag, src_data)

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickupAndCompareForUpdate(src_data, curr_data, self.properties_for_update)
        if update_data != None:
            self._update(confirmation_flag, update_data)

    properties = [
        "RoleArn",
        "Type",
        "LoggingConfiguration",
        "TracingConfiguration",
    ]
    properties_for_update = [
        "RoleArn",
        "LoggingConfiguration",
        "TracingConfiguration",
    ]

class StateMachineDefinitionHandler(StateMachineBasicHandler):
    def __init__(self, session, stepfunctions_client, machine_name):
        StateMachineBasicHandler.__init__(self, session, stepfunctions_client, machine_name)

    def describe(self):
        info = self._describe()
        if info == None:
            return None
        result = self._definition_str_to_dict(info["Definition"])
        if "__skwadon_dummy" in result["States"]:
            del result["States"]["__skwadon_dummy"]
        return result

    def put(self, confirmation_flag, src_data):
        update_data = {
            "Definition": self._definition_dict_to_str(src_data),
        }
        self._update(confirmation_flag, update_data)

    def _definition_str_to_dict(self, definition: str):
        return json.loads(definition)

    def _definition_dict_to_str(self, definition: dict):
        return json.dumps(definition)

def calc_statemachine_arn(session, machine_name):
    account_id = sic_aws.fetch_account_id(session)
    region_name = sic_aws.fetch_region_name(session)
    return calc_statemachine_arn2(account_id, region_name, machine_name)

def calc_statemachine_arn2(account_id, region_name, machine_name):
    return f"arn:aws:states:{region_name}:{account_id}:stateMachine:{machine_name}"



from abc import abstractmethod

import skwadon.action_lib as action_lib

# Handlerは各階層でのdo_get, do_putを実装するクラス。
# 各階層ごとに再帰的に処理していく。
#
# 階層とは
#   $ skwadon aws iam.roles.AWSServiceRoleForAPIGateway
# と実行した場合
#   - iam
#   - iam.roles
#   - iam.roles.AWSServiceRoleForAPIGateway
#  の3階層。
#
# このファイルではHandlerのほかにHandlerのサブクラスを定義している。
#
# ListHandler
#   リソースの一覧を表現する階層
#   AWSの各サービスでのリソース一覧はこのクラスのサブクラスを実装する
# NamespaceHandler
#   名前を区切っているだけの階層
#   サブクラスを実装することなくAWSの各サービスで直接インスタンスを生成する
# ResourceHandler
#   リソースの実際の属性値などを表現する階層
#   AWSの各サービスでこのクラスのサブクラスを実装する
# ResourceInfoHandler
#   describeのみを実装しているResourceHandlerのサブクラス


class Handler:

    # getアクションを実装
    @abstractmethod
    def do_get(self, src_data, thats_all_flag):
        return {}

    # putアクションを実装 (deleteアクションを含む)
    # 返り値
    #   dryrun(confirmation_flag=False): (入力, 変更前のクラウド)
    #   更新実行(confirmation_flag=True): (変更後のクラウド, 変更前のクラウド)
    @abstractmethod
    def do_put(self, confirmation_flag, src_data):
        return ({}, {})


# abstract
class ListHandler(Handler):

    def do_get(self, src_data, thats_all_flag):
        if hasattr(self, "is_default_all") and self.is_default_all:
            default_keys = None
        else:
            default_keys = []
        return action_lib.encode_for_get_list(
            src_data, thats_all_flag, default_keys,
            lambda: self._list(),
            lambda name, src_data: self.child_handler(name).do_get(src_data, thats_all_flag),
        )

    def do_put(self, confirmation_flag, src_data):
        if src_data is None:
            src_data = {"*": None}
        elif not isinstance(src_data, dict):
            src_data = {}
        result1 = {}  # 入力または変更後のクラウド
        result2 = {}  # 変更前のクラウド
        for name, value in src_data.items():
            if name == "*":
                for name2 in self._list():
                    if name2 not in src_data:
                        d1, d2 = self.child_handler(name2).do_put(confirmation_flag, None)
                        result1[name2] = None
                        result2[name2] = d2
            elif value is None:
                curr_data = self.child_handler(name).do_get(None, False)
                if curr_data is not None:
                    d1, d2 = self.child_handler(name).do_put(confirmation_flag, None)
                    result1[name] = None
                    result2[name] = d2
            else:
                d1, d2 = self.child_handler(name).do_put(confirmation_flag, value)
                result1[name] = d1
                result2[name] = d2
        return (result1, result2)

    def _list(self):
        return self.sort_items(self.list())

    @abstractmethod
    def list(self):
        raise Exception(f"list method not implemented: {self}")

    @abstractmethod
    def child_handler(self, name):
        return Handler()

    def sort_items(self, list):
        return sorted(list)


class NamespaceHandler(Handler):
    def __init__(self, conf_key, default_keys, path_map):
        self.conf_key = conf_key
        self.default_keys = default_keys
        handler_map = {}
        for path, handler in path_map.items():
            p = path.find(".")
            if p < 0:
                h = path
                t = ""
            else:
                h = path[0:p]
                t = path[p+1:]
            if h not in handler_map:
                handler_map[h] = {}
            handler_map[h][t] = handler
        self.handler_map = {}
        for key, map in handler_map.items():
            if "" in map:
                self.handler_map[key] = map[""]
            else:
                self.handler_map[key] = NamespaceHandler(None, [], map)

    def do_get(self, src_data, thats_all_flag):
        if self.conf_key:
            conf = self.handler_map[self.conf_key].do_get(None, False)
            if conf is None:
                return None
        def sub(name, src_data):
            if name in self.handler_map:
                return self.handler_map[name].do_get(src_data, thats_all_flag)
            else:
                return None
        result = action_lib.encode_for_get_list(
            src_data, thats_all_flag, self.default_keys,
            lambda: self.handler_map.keys(),
            lambda name, src_data: sub(name, src_data),
        )
        return result

    def do_put(self, confirmation_flag, src_data):
        if self.conf_key:
            conf = self.handler_map[self.conf_key].do_get(None, False)
            if conf is None:
                if src_data is None:
                    return (None, None)
        if src_data is None:
            src_data = {"*": None}
        elif not isinstance(src_data, dict):
            src_data = {}
        result1 = {}  # 入力または変更後のクラウド
        result2 = {}  # 変更前のクラウド
        for name, value in src_data.items():
            if name == "*":
                if value is None:
                    result2r = {}
                    for name2 in reversed(self.handler_map.keys()):
                        if name2 not in src_data:
                            d1, d2 = self.handler_map[name2].do_put(confirmation_flag, None)
                            result2r[name2] = d2
                    for name2 in reversed(result2r.keys()):
                        result1[name2] = None
                        result2[name2] = result2r[name2]
                result1[name] = value
            elif value is None:
                if name in self.handler_map:
                    d1, d2 = self.handler_map[name].do_put(confirmation_flag, None)
                    result1[name] = None
                    result2[name] = d2
            else:
                if name in self.handler_map:
                    d1, d2 = self.handler_map[name].do_put(confirmation_flag, value)
                    result1[name] = d1
                    result2[name] = d2
        return (result1, result2)


# abstract
class ResourceHandler(Handler):
    def do_get(self, src_data, thats_all_flag):
        curr_data = self.describe()
        return action_lib.encode_for_get_resource2(src_data, curr_data, thats_all_flag)

    def do_put(self, confirmation_flag, src_data):
        curr_data = self.describe()

        # 1つ目の返り値: 実際に self.updateに渡す値
        # 2つ目の返り値: コマンドの出力結果としての入力
        # 3つ目の返り値: コマンドの出力結果としての変更前のクラウド
        def build_update_data(src_data, curr_data):
            if not isinstance(src_data, dict) or not isinstance(curr_data, dict):
                src_data2 = action_lib.decode_for_put(src_data)
                result2 = action_lib.encode_for_get_resource(curr_data)
                return (src_data2, src_data, result2)
            src_data2 = {}  # 実際に self.updateに渡す値
            result1 = {}
            result2 = {}
            delete_flag = False
            for name, value in src_data.items():
                if name == "*":
                    if value is None:
                        for name2 in curr_data:
                            if name2 not in src_data:
                                result2[name2] = action_lib.encode_for_get_resource(curr_data[name2])
                        delete_flag = True
                elif value is None:
                    if name in curr_data:
                        result2[name] = action_lib.encode_for_get_resource(curr_data[name])
                else:
                    if name in curr_data:
                        src_data2[name], result1[name], result2[name] = build_update_data(value, curr_data[name])
                    else:
                        src_data2[name], result1[name], _ = build_update_data(value, None)
            if not delete_flag:
                for name in curr_data:
                    if name not in src_data2:
                        src_data2[name] = action_lib.encode_for_get_resource(curr_data[name])
            return (src_data2, result1, result2)

        src_data2, result1, result2 = build_update_data(src_data, curr_data)
        if src_data2 == curr_data:
            return (result1, result2)

        if curr_data is None:
            self.create(confirmation_flag, src_data2)
        elif src_data is None:
            self.delete(confirmation_flag, curr_data)
        else:
            self.update(confirmation_flag, src_data2, curr_data)

        if confirmation_flag:
            _, _, result1 = build_update_data(src_data, self.describe())

        return (result1, result2)

    @abstractmethod
    def describe(self):
        raise Exception(f"describe method not implemented: {self}")

    @abstractmethod
    def create(self, confirmation_flag, src_data):
        self.put(confirmation_flag, src_data)

    @abstractmethod
    def update(self, confirmation_flag, src_data, curr_data):
        self.put(confirmation_flag, src_data)

    @abstractmethod
    def put(self, confirmation_flag, src_data):
        raise Exception(f"put method not implemented: {self}")

    @abstractmethod
    def delete(self, confirmation_flag, curr_data):
        raise Exception(f"delete method not implemented: {self}")


class ResourceInfoHandler(ResourceHandler):

    def __init__(self, info):
        self.info = info

    def describe(self):
        return self.info




import json

import botocore.exceptions

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as aws
import skwadon.aws_iam_policy as iam_policy

class RoleListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.iam_client = None

    def init_client(self):
        if self.iam_client == None:
            self.iam_client = self.session.client("iam")
            self.account_id = aws.fetch_account_id(self.session)

    def list(self):
        self.init_client()
        result = []
        res = self.iam_client.list_roles()
        while True:
            for elem in res["Roles"]:
                name = elem["RoleName"]
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_roles(Marker = res["Marker"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "inlinePolicies", "attachedPolicies", "assumeRolePolicy"], {
            "conf": RoleConfHandler(self.iam_client, name),
            "inlinePolicies": RoleInlinePolicyListHandler(self.iam_client, name),
            "attachedPolicies": AttachedPolicyListHandler(self.iam_client, name, self.account_id),
            "assumeRolePolicy": AssumeRolePolicyHandler(self.iam_client, name),
        })

class RoleConfHandler(common_action.ResourceHandler):
    properties = [
        "Path",
        "RoleId",
        "Description",
        "MaxSessionDuration",
    ]

    def __init__(self, iam_client, role_name):
        self.iam_client = iam_client
        self.role_name = role_name

    def describe(self):
        try:
            res = self.iam_client.get_role(RoleName = self.role_name)
            curr_data = sic_lib.pickup(res["Role"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = sic_lib.pickup(src_data, ["Description", "MaxSessionDuration"])
        update_data["RoleName"] = self.role_name
        update_data["AssumeRolePolicyDocument"] = json.dumps(self._defaultAssumeRole())
        sic_main.exec_put(confirmation_flag,
            f"iam_client.create_role(RoleName = {self.role_name}, ...)",
            lambda: self.iam_client.create_role(**update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickupAndCompareForUpdate(src_data, curr_data, ["Description", "MaxSessionDuration"])
        if update_data != None:
            update_data["RoleName"] = self.role_name
            sic_main.exec_put(confirmation_flag,
                f"iam_client.update_role(RoleName = {self.role_name}, ...)",
                lambda:
                    self.iam_client.update_role(**update_data)
            )

    def delete(self, confirmation_flag, curr_data):
        sic_main.exec_put(confirmation_flag,
            f"iam_client.delete_role(RoleName = {self.role_name})",
            lambda:
                self.iam_client.delete_role(RoleName = self.role_name)
        )

    def _defaultAssumeRole(self):
        return {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {
                        "Service": "ec2.amazonaws.com"
                    },
                    "Action": "sts:AssumeRole"
                }
            ],
        }

class RoleInlinePolicyListHandler(common_action.ListHandler):
    def __init__(self, iam_client, role_name):
        self.is_default_all = True
        self.iam_client = iam_client
        self.role_name = role_name

    def list(self):
        result = []
        res = self.iam_client.list_role_policies(RoleName = self.role_name)
        while True:
            for elem in res["PolicyNames"]:
                name = elem
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_role_policies(RoleName = self.role_name, Marker = res["Marker"])
        return result

    def child_handler(self, name):
        return RoleInlinePolicyHandler(self.iam_client, self.role_name, name)

class RoleInlinePolicyHandler(common_action.ResourceHandler):
    def __init__(self, iam_client, role_name, policy_name):
        self.iam_client = iam_client
        self.role_name = role_name
        self.policy_name = policy_name

    def describe(self):
        try:
            res = self.iam_client.get_role_policy(RoleName = self.role_name, PolicyName = self.policy_name)
            return res["PolicyDocument"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

    def put(self, confirmation_flag, src_data):
        policy_document_json = json.dumps(src_data)
        sic_main.exec_put(confirmation_flag,
            f"iam_client.put_role_policy(RoleName = {self.role_name}, PolicyName = {self.policy_name}, ...)",
            lambda:
                self.iam_client.put_role_policy(RoleName = self.role_name, PolicyName = self.policy_name, PolicyDocument = policy_document_json)
        )

    def delete(self, confirmation_flag, curr_data):
        sic_main.exec_put(confirmation_flag,
            f"iam_client.delete_role_policy(RoleName = {self.role_name}, PolicyName = {self.policy_name})",
            lambda:
                self.iam_client.delete_role_policy(RoleName = self.role_name, PolicyName = self.policy_name)
        )

class AttachedPolicyListHandler(common_action.ListHandler):
    def __init__(self, iam_client, role_name, account_id):
        self.iam_client = iam_client
        self.role_name = role_name
        self.account_id = account_id

    def list(self):
        result = []
        res = self.iam_client.list_attached_role_policies(RoleName = self.role_name)
        while True:
            for elem in res["AttachedPolicies"]:
                name = iam_policy.policy_arn_to_name(elem["PolicyArn"], self.account_id)
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_role_policies(RoleName = self.role_name, Marker = res["Marker"])
        return result

class AssumeRolePolicyHandler(common_action.ResourceHandler):
    def __init__(self, iam_client, role_name):
        self.iam_client = iam_client
        self.role_name = role_name

    def describe(self):
        try:
            res = self.iam_client.get_role(RoleName = self.role_name)
            curr_data = sic_lib.dictlib_get(res, "Role.AssumeRolePolicyDocument")
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

    def put(self, confirmation_flag, src_data):
        policy_document_json = json.dumps(src_data)
        sic_main.exec_put(confirmation_flag,
            f"iam_client.update_assume_role_policy(RoleName = {self.role_name}, ...)",
            lambda:
                self.iam_client.update_assume_role_policy(RoleName = self.role_name, PolicyDocument = policy_document_json)
        )

    def delete(self, confirmation_flag, curr_data):
        pass




def encode_for_get_list(src_data, thats_all_flag, default_keys, items_fetcher, item_detail_fetcher):
    if not isinstance(src_data, dict):
        src_data = {"*": None}
    result = {}
    unknown_flag = False
    for name, value in src_data.items():
        if name == "*":
            for name2 in items_fetcher():
                if not name2 in src_data:
                    if value != None and value != {}:
                        result[name2] = item_detail_fetcher(name2, value)
                    elif value == None and (default_keys is None or name2 in default_keys):
                        result[name2] = item_detail_fetcher(name2, None)
                    else:
                        result[name2] = {}
            unknown_flag = True
        else:
            d = item_detail_fetcher(name, value)
            if d != None:
                result[name] = d
    if unknown_flag and thats_all_flag:
        result["*"] = None
    return result

def encode_for_get_resource(data):
    return encode_for_get_resource2(None, data, False)

def encode_for_get_resource2(src_data, curr_data, thats_all_flag):
    if not isinstance(curr_data, dict):
        return curr_data
    if not isinstance(src_data, dict):
        src_data = {"*": None}
    result = {}
    unknown_flag = False
    for name, value in src_data.items():
        if name == "*":
            if value == None:
                for name2, value2 in curr_data.items():
                    if not name2 in src_data:
                        result[name2] = encode_for_get_resource2(value, value2, thats_all_flag)
                unknown_flag = True
        elif name in curr_data:
            result[name] = encode_for_get_resource2(value, curr_data[name], thats_all_flag)
        else:
            pass
    # AccessDeniedの場合は curr_data に {"*": "AccessDenied"} が来る
    # その場合は thats_all_flag があっても {"*": null} にしない
    if unknown_flag and thats_all_flag and "*" not in curr_data:
        result["*"] = None
    return result

def decode_for_put(src_data):
    if isinstance(src_data, str):
        return src_data
    if not isinstance(src_data, dict):
        return src_data
    src_data2 = {}
    for name, value in src_data.items():
        if name == "*":
            continue
        src_data2[name] = decode_for_put(value)
    return src_data2






import yaml

import skwadon.main as sic_main
import skwadon.common_action as common_action

def get_message_prefix(data):
    return "testcloud"

def get_handler(src_data):
    return common_action.NamespaceHandler(
        None, [], {
        "seeds": SeedListHandler(),
    })

class SeedListHandler(common_action.ListHandler):

    _items = [
        "himawari1",
        "himawari2",
        "himawari3",
    ]

    def list(self):
        return self._items

    def child_handler(self, name):
        exists = name in self._items
        return common_action.NamespaceHandler(
            "conf", ["conf"], {
            "conf": SeedConfHandler(name, exists),
            "status": SeedStatusHandler(name, exists),
        })

class SeedConfHandler(common_action.ResourceHandler):
    def __init__(self, name, exists):
        self.name = name
        self.exists = exists

    def describe(self):
        if not self.exists:
            return None
        return {
            "Description": f"himawari no tane '{self.name}'",
            "Count": 10,
            "Color": {
                "r": 30,
                "g": 20,
                "b": 15,
            },
        }

    def create(self, confirmation_flag, src_data):
        print(f"TESTCLOUD: create conf, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"src_data": src_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: create conf",
            lambda:
                print("TESTCLOUD: execute create conf"),
        )

    def update(self, confirmation_flag, src_data, curr_data):
        print(f"TESTCLOUD: update conf, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"src_data": src_data, "curr_data": curr_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: update conf",
            lambda:
                print("TESTCLOUD: execute update conf"),
        )

    def delete(self, confirmation_flag, curr_data):
        print(f"TESTCLOUD: delete, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"curr_data": curr_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: delete conf",
            lambda:
                print("TESTCLOUD: execute delete conf"),
        )

class SeedStatusHandler(common_action.ResourceHandler):
    def __init__(self, name, exists):
        self.name = name
        self.exists = exists

    def describe(self):
        if not self.exists:
            return None
        return {
            "status": "OK",
        }

    def create(self, confirmation_flag, src_data):
        print(f"TESTCLOUD: create status, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"src_data": src_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: create status",
            lambda:
                print("TESTCLOUD: execute create status"),
        )

    def update(self, confirmation_flag, src_data, curr_data):
        print(f"TESTCLOUD: update status, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"src_data": src_data, "curr_data": curr_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: update status",
            lambda:
                print("TESTCLOUD: execute update status"),
        )

    def delete(self, confirmation_flag, curr_data):
        print(f"TESTCLOUD: delete status, confirmation_flag = {confirmation_flag}")
        sic_main.save_yaml({"curr_data": curr_data}, None)
        sic_main.exec_put(
            confirmation_flag,
            "api: delete status",
            lambda:
                print("TESTCLOUD: execute delete status"),
        )




import copy
import json

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action

class CrawlerListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.glue_client = None

    def init_client(self):
        if self.glue_client == None:
            self.glue_client = self.session.client("glue")

    def list(self):
        self.init_client()
        result = []
        res = self.glue_client.get_crawlers()
        while True:
            for elem in res['Crawlers']:
                name = elem["Name"]
                result.append(name)
            if not "NextToken" in res:
                break
            res = self.glue_client.get_crawlers(NextToken = res["NextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf"], {
            "conf": CrawlerConfHandler(self.glue_client, name),
            "status": CrawlerStatusHandler(self.glue_client, name),
        })

class CrawlerConfHandler(common_action.ResourceHandler):

    properties = [
        "Role",
        "Targets",
        "DatabaseName",
        "Description",
        "Classifiers",
        "RecrawlPolicy",
        "SchemaChangePolicy",
        "LineageConfiguration",
        "TablePrefix",
        "Schedule",
        "Configuration",
        "CrawlerSecurityConfiguration",
        "LakeFormationConfiguration",
    ]

    def __init__(self, glue_client, crawler_name):
        self.glue_client = glue_client
        self.crawler_name = crawler_name

    def describe(self):
        try:
            res = self.glue_client.get_crawler(Name = self.crawler_name)
            curr_data = sic_lib.pickup(res["Crawler"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = sic_lib.pickup(src_data, self.properties)
        update_data["Name"] = self.crawler_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_crawler(Name = {self.crawler_name}, ...)",
            lambda:
                self.glue_client.create_crawler(**update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickupAndCompareForUpdate(src_data, curr_data, self.properties)
        if update_data != None:
            update_data["Name"] = self.crawler_name
            sic_main.exec_put(confirmation_flag,
                f"glue_client.update_crawler(Name = {self.crawler_name}, ...)",
                lambda:
                    self.glue_client.update_crawler(**update_data)
            )

    def delete(self, confirmation_flag, curr_data):
        sic_main.add_update_message(f"glue_client.delete_crawler(Name = {self.crawler_name})")
        if confirmation_flag and sic_main.global_confirmation_flag:
            self.glue_client.delete_crawler(Name = self.crawler_name)

class CrawlerStatusHandler(common_action.ResourceHandler):

    properties = [
        "State",
        "CrawlElapsedTime",
        "CreationTime",
        "LastUpdated",
        "LastCrawl",
        "Version",
    ]

    def __init__(self, glue_client, crawler_name):
        self.glue_client = glue_client
        self.crawler_name = crawler_name

    def describe(self):
        res = self.glue_client.get_crawler(Name = self.crawler_name)
        curr_data = sic_lib.pickup(res["Crawler"], self.properties)
        return curr_data

    def delete(self, confirmation_flag, curr_data):
        pass



import copy
import shutil
import tempfile
import time
import urllib.request

import botocore.exceptions

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as sic_aws

class FunctionListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.lambda_client = None

    def init_client(self):
        if self.lambda_client == None:
            self.lambda_client = self.session.client("lambda")

    def list(self):
        self.init_client()
        result = []
        res = self.lambda_client.list_functions()
        while True:
            for elem in res['Functions']:
                name = elem["FunctionName"]
                result.append(name)
            if not "NextMarker" in res:
                break
            res = self.lambda_client.list_functions(NextMarker = res["NextMarker"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "sources"], {
            "conf": FunctionConfHandler(self.lambda_client, name),
            "sources": FunctionSourcesHandler(self.lambda_client, name),
        })

class FunctionConfHandler(common_action.ResourceHandler):

    properties = [
        'Runtime',
        'Role',
        'Handler',
        'Description',
        'Timeout',
        'MemorySize',
        'VpcConfig',
        'DeadLetterConfig',
        'Environment',
        'KMSKeyArn',
        'TracingConfig',
        'MasterArn',
        'Layers',
        'FileSystemConfigs',
        'PackageType',
        'ImageConfigResponse',
        'Architectures',
    ]

    def __init__(self, lambda_client, function_name):
        self.lambda_client = lambda_client
        self.function_name = function_name

    def describe(self):
        try:
            res = self.lambda_client.get_function(FunctionName = self.function_name)
            curr_data = sic_lib.pickup(res["Configuration"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = src_data.copy()
        update_data["FunctionName"] = self.function_name
        update_data["Code"] = {"ZipFile": create_zip({"__skwadon_dummy.txt": "dummy"})}
        sic_main.exec_put(confirmation_flag,
            f"lambda_client.create_function(FunctionName = {self.function_name}, ...)",
            lambda: self.lambda_client.create_function(**update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = src_data.copy()
        update_data["FunctionName"] = self.function_name
        sic_lib.removeKey(update_data, "PackageType")
        sic_lib.removeKey(update_data, "Architectures")
        sic_main.exec_put(confirmation_flag,
            f"lambda_client.update_function_configuration(FunctionName = {self.function_name}, ...)",
            lambda: self.lambda_client.update_function_configuration(**update_data)
        )


class FunctionSourcesHandler(common_action.ResourceHandler):

    def __init__(self, lambda_client, function_name):
        self.lambda_client = lambda_client
        self.function_name = function_name

    def describe(self):
        try:
            res = self.lambda_client.get_function(FunctionName = self.function_name)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return None
            else:
                raise
        result = {}
        if res["Configuration"]["PackageType"] == "Zip" and res["Code"]["RepositoryType"] == "S3":
            url = res["Code"]["Location"]
            with tempfile.TemporaryDirectory() as tmp_dir:
                with tempfile.NamedTemporaryFile() as tmp_file:
                    with urllib.request.urlopen(url) as response:
                        shutil.copyfileobj(response, tmp_file)
                    tmp_file.flush()
                    shutil.unpack_archive(tmp_file.name, tmp_dir, "zip")
                result = sic_lib.script_sources_to_yaml(tmp_dir)
        return result

    def put(self, confirmation_flag, src_data):
        if confirmation_flag:
            self.waitPending()
        src_data2 = src_data.copy()
        if "__skwadon_dummy.txt" in src_data2:
            del src_data2["__skwadon_dummy.txt"]
        zipbin = create_zip(src_data2)
        update_data = {}
        update_data["FunctionName"] = self.function_name
        update_data["ZipFile"] = zipbin
        sic_main.exec_put(confirmation_flag,
            f"lambda_client.update_function_code(FunctionName = {self.function_name}, ...)",
            lambda: self.lambda_client.update_function_code(**update_data)
        )

    def waitPending(self):
        while True:
            res = self.lambda_client.get_function(FunctionName = self.function_name)
            if res["Configuration"]["State"] == "Pending":
                time.sleep(1)
            else:
                break

def create_zip(sources):
    with tempfile.TemporaryDirectory() as tmp_dir:
        sic_lib.script_sources_from_yaml(tmp_dir, sources)
        with tempfile.TemporaryDirectory() as tmp_dir2:
            shutil.make_archive(tmp_dir2 + "/sources", "zip", tmp_dir)
            with open(tmp_dir2 + "/sources.zip", "rb") as fh:
                zipbin = fh.read()
    return zipbin




import json

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action


class BucketListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.s3_client = None

    def init_client(self):
        if self.s3_client is None:
            self.s3_client = self.session.client("s3")
            self.s3_resource = self.session.resource("s3")

    def list(self):
        self.init_client()
        result = []
        res = self.s3_client.list_buckets()
        for elem in res['Buckets']:
            name = elem["Name"]
            result.append(name)
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "bucketPolicy", "lifecycles"],
            {
                "conf": BucketConfHandler(self.s3_client, name),
                "bucketPolicy": BucketPolicyHandler(self.s3_client, name),
                "lifecycles": BucketLifecyclesHandler(self.s3_client, name),
            },
        )


class BucketConfHandler(common_action.ResourceHandler):

    def __init__(self, s3_client, bucket_name):
        self.s3_client = s3_client
        self.bucket_name = bucket_name

    def describe(self):
        curr_data = {}

        try:
            res = self.s3_client.get_bucket_location(Bucket=self.bucket_name)
            curr_data["location"] = res["LocationConstraint"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchBucket":
                return None
            elif e.response["Error"]["Code"] == "AccessDenied":
                return {"*": "AccessDenied"}
            else:
                raise

        try:
            res = self.s3_client.get_bucket_policy_status(Bucket=self.bucket_name)
            curr_data["publicAccessBlock"] = not res["PolicyStatus"]["IsPublic"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchBucketPolicy":
                curr_data["publicAccessBlock"] = False
            elif e.response["Error"]["Code"] == "AccessDenied":
                return {"*": "AccessDenied"}
            else:
                raise

        return curr_data


class BucketPolicyHandler(common_action.ResourceHandler):
    def __init__(self, s3_client, bucket_name):
        self.s3_client = s3_client
        self.bucket_name = bucket_name

    def describe(self):
        try:
            res = self.s3_client.get_bucket_policy(Bucket=self.bucket_name)
            curr_data = json.loads(res["Policy"])
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchBucketPolicy":
                return None
            raise


class BucketLifecyclesHandler(common_action.ResourceHandler):
    def __init__(self, s3_client, bucket_name):
        self.s3_client = s3_client
        self.bucket_name = bucket_name

    def describe(self):
        try:
            res = self.s3_client.get_bucket_lifecycle_configuration(Bucket=self.bucket_name)
            result = self._encode(res["Rules"])
            return result
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchLifecycleConfiguration":
                return {}
            elif e.response["Error"]["Code"] == "AccessDenied":
                return {"*": "AccessDenied"}
            raise

    def put(self, confirmation_flag, src_data):
        update_data = {"Rules": self._decode(src_data)}
        if len(update_data["Rules"]) == 0:
            sic_main.exec_put(
                confirmation_flag,
                f"s3_client.delete_bucket_lifecycle(Bucket={self.bucket_name})",
                lambda:
                    self.s3_client.delete_bucket_lifecycle(Bucket=self.bucket_name)
            )
        else:
            sic_main.exec_put(
                confirmation_flag,
                f"s3_client.put_bucket_lifecycle_configuration(Bucket={self.bucket_name}, LifecycleConfiguration=...)",
                lambda:
                    self.s3_client.put_bucket_lifecycle_configuration(Bucket=self.bucket_name, LifecycleConfiguration=update_data)
            )

    def delete(self, confirmation_flag, curr_data):
        self.put(confirmation_flag, None)

    def _encode(self, info):
        result = {}
        for elem in info:
            elem2 = elem.copy()
            name = sic_lib.encode_key(elem["ID"])
            del elem2["ID"]
            if "Prefix" not in elem2:
                elem2["Prefix"] = ""
            result[name] = elem2
        return result

    def _decode(self, info):
        if info is None:
            return []
        result = []
        for name, elem in info.items():
            elem2 = elem.copy()
            elem2["ID"] = sic_lib.decode_key(name)

            # putのときは Prefix が必須

            result.append(elem2)
        return result





import copy

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action

class TriggerListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.glue_client = None

    def init_client(self):
        if self.glue_client is None:
            self.glue_client = self.session.client("glue")

    def list(self):
        self.init_client()
        result = []
        res = self.glue_client.get_triggers()
        while True:
            for elem in res['Triggers']:
                name = elem["Name"]
                result.append(name)
            if "NextToken" not in res:
                break
            res = self.glue_client.get_databases(NextToken=res["NextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "status"], {
            "conf": TriggerConfHandler(self.glue_client, name),
            "status": TriggerStatusHandler(self.glue_client, name),
        })

class TriggerConfHandler(common_action.ResourceHandler):

    properties = [
        "Type",
        "Description",
        "Schedule",
        "Actions",
    ]
    properties_for_update = [
        "Description",
        "Schedule",
        "Actions",
    ]

    def __init__(self, glue_client, trigger_name):
        self.glue_client = glue_client
        self.trigger_name = trigger_name

    def describe(self):
        try:
            res = self.glue_client.get_trigger(Name = self.trigger_name)
            curr_data = sic_lib.pickup(res["Trigger"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = copy.copy(src_data)
        update_data["Name"] = self.trigger_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_trigger({{Name: {self.trigger_name}, ...}})",
            lambda:
                self.glue_client.create_trigger(**update_data)
        )
        sic_main.exec_put(confirmation_flag,
            f"glue_client.start_trigger(Name={self.trigger_name})",
            lambda:
                self.glue_client.start_trigger(Name=self.trigger_name)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickup(src_data, self.properties_for_update)
        sic_main.exec_put(confirmation_flag,
            f"glue_client.update_trigger(Name={self.trigger_name}, TriggerUpdate={{...}})",
            lambda:
                self.glue_client.update_trigger(Name=self.trigger_name,
                TriggerUpdate=update_data)
        )

    def delete(self, confirmation_flag, curr_data):
        sic_main.exec_put(confirmation_flag,
            f"glue_client.delete_trigger(Name={self.trigger_name})",
            lambda:
                self.glue_client.delete_trigger(Name=self.trigger_name)
        )

class TriggerStatusHandler(common_action.ResourceHandler):

    properties = [
        "State",
    ]

    def __init__(self, glue_client, trigger_name):
        self.glue_client = glue_client
        self.trigger_name = trigger_name

    def describe(self):
        try:
            res = self.glue_client.get_trigger(Name = self.trigger_name)
            curr_data = sic_lib.pickup(res["Trigger"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def delete(self, confirmation_flag, curr_data):
        pass




import json

import botocore.exceptions

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as aws
import skwadon.aws_iam_policy as iam_policy

class GroupListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.iam_client = None

    def init_client(self):
        if self.iam_client == None:
            self.iam_client = self.session.client("iam")
            self.account_id = aws.fetch_account_id(self.session)

    def list(self):
        self.init_client()
        result = []
        res = self.iam_client.list_groups()
        while True:
            for elem in res["Groups"]:
                name = elem["GroupName"]
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_groups(Marker = res["Marker"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf", "inlinePolicies", "attachedPolicies"], {
                "conf": GroupConfHandler(self.iam_client, name),
                "inlinePolicies": GroupInlinePolicyListHandler(self.iam_client, name),
                "attachedPolicies": AttachedPolicyListHandler(self.iam_client, name, self.account_id),
            },
        )

class GroupConfHandler(common_action.ResourceHandler):
    properties = [
        "Path",
        "GroupId",
    ]

    def __init__(self, iam_client, name):
        self.iam_client = iam_client
        self.name = name

    def describe(self):
        try:
            res = self.iam_client.get_group(GroupName=self.name)
            curr_data = sic_lib.pickup(res["Group"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

class GroupInlinePolicyListHandler(common_action.ListHandler):
    def __init__(self, iam_client, group_name):
        self.is_default_all = True
        self.iam_client = iam_client
        self.group_name = group_name

    def list(self):
        result = []
        res = self.iam_client.list_group_policies(GroupName = self.group_name)
        while True:
            for elem in res["PolicyNames"]:
                name = elem
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_group_policies(GroupName = self.role_name, Marker = res["Marker"])
        return result

    def child_handler(self, name):
        return GroupInlinePolicyHandler(self.iam_client, self.group_name, name)

class GroupInlinePolicyHandler(common_action.ResourceHandler):
    def __init__(self, iam_client, group_name, policy_name):
        self.iam_client = iam_client
        self.group_name = group_name
        self.policy_name = policy_name

    def describe(self):
        try:
            res = self.iam_client.get_group_policy(GroupName = self.group_name, PolicyName = self.policy_name)
            return res["PolicyDocument"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

class AttachedPolicyListHandler(common_action.ListHandler):
    def __init__(self, iam_client, group_name, account_id):
        self.iam_client = iam_client
        self.group_name = group_name
        self.account_id = account_id

    def list(self):
        result = []
        res = self.iam_client.list_attached_group_policies(GroupName=self.group_name)
        while True:
            for elem in res["AttachedPolicies"]:
                name = iam_policy.policy_arn_to_name(elem["PolicyArn"], self.account_id)
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_attached_group_policies(GroupName=self.group_name, Marker=res["Marker"])
        return result





import skwadon.lib as sic_lib
import skwadon.common_action as common_action


class ClusterListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.redshift_client = None
        self.clusters = None

    def init_client(self):
        if self.redshift_client is None:
            self.redshift_client = self.session.client("redshift")
            self.clusters = {}
            res = self.redshift_client.describe_clusters()
            while True:
                for elem in res['Clusters']:
                    name = elem["ClusterIdentifier"]
                    self.clusters[name] = elem
                if "Marker" not in res:
                    break
                res = self.redshift_client.describe_clusters(Marker=res["Marker"])

    def list(self):
        self.init_client()
        result = []
        for name in self.clusters:
            result.append(name)
        return result

    def child_handler(self, name):
        self.init_client()
        if name in self.clusters:
            info = self.clusters[name]
        else:
            info = None
        info = self._create_instance_info(info)
        return common_action.NamespaceHandler(
            "conf", ["conf"],
            {
                "conf":       common_action.ResourceInfoHandler(info["conf"]),
                "all":        common_action.ResourceInfoHandler(info["all"]),
                "connection": ClusterConnectionHandler(self.redshift_client, name, info["all"]),
            },
        )

    def _create_instance_info(self, info):
        return {
            "conf": sic_lib.pickup(info, [
                "NodeType",
                "NumberOfNodes",
                "MasterUsername",
                "DBName",
            ]),
            "all": info,
        }


class ClusterConnectionHandler(common_action.ResourceHandler):

    def __init__(self, redshift_client, cluster_name, cluster_info):
        self.redshift_client = redshift_client
        self.cluster_name = cluster_name
        self.cluster_info = cluster_info

    def describe(self):
        request_info = {
            "DbUser": self.cluster_info["MasterUsername"],
            "DbName": self.cluster_info["DBName"],
            "ClusterIdentifier": self.cluster_name,
            "DurationSeconds": 3600,
            "AutoCreate": False,
        }
        res = self.redshift_client.get_cluster_credentials(**request_info)
        cmd = f"PGPASSWORD='{res['DbPassword']}' psql -h {self.cluster_info['Endpoint']['Address']} -p {self.cluster_info['Endpoint']['Port']} -U {res['DbUser']} -d {self.cluster_info['DBName']}"
        result = {
            "Endpoint": {
                "Address": self.cluster_info["Endpoint"]["Address"],
                "Port": self.cluster_info["Endpoint"]["Port"],
            },
            "DbName": self.cluster_info["DBName"],
            "DbUser": res["DbUser"],
            "DbPassword": res["DbPassword"],
            "Expiration": res["Expiration"],
            "CommandLine": cmd,
        }
        return result






import skwadon.lib as sic_lib
import skwadon.common_action as common_action


class InstanceListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.rds_client = None
        self.instances = None

    def init_client(self):
        if self.rds_client is None:
            self.rds_client = self.session.client("rds")
            self.instances = {}
            res = self.rds_client.describe_db_instances()
            while True:
                for elem in res['DBInstances']:
                    name = elem["DBInstanceIdentifier"]
                    self.instances[name] = elem
                if "Marker" not in res:
                    break
                res = self.rds_client.describe_db_instances(Marker=res["Marker"])

    def list(self):
        self.init_client()
        result = []
        for name in self.instances:
            result.append(name)
        return result

    def child_handler(self, name):
        self.init_client()
        if name in self.instances:
            info = self.instances[name]
        else:
            info = None
        info = self._create_instance_info(info)
        return common_action.NamespaceHandler(
            "conf", ["conf"],
            {
                "conf":   common_action.ResourceInfoHandler(info["conf"]),
                "status": common_action.ResourceInfoHandler(info["status"]),
                "all":    common_action.ResourceInfoHandler(info["all"]),
            },
        )

    def _create_instance_info(self, info):
        return {
            "conf": sic_lib.pickup(info, [
                "DBInstanceClass",
                "Engine",
                "MasterUsername",
                "DBName",
            ]),
            "status": sic_lib.pickup(info, [
                "DBInstanceStatus",
                "AutomaticRestartTime",
            ]),
            "all": info,
        }




import json
import re

import botocore.exceptions

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as aws

class PolicyListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.iam_client = None

    def init_client(self):
        if self.iam_client == None:
            self.iam_client = self.session.client("iam")
            self.account_id = aws.fetch_account_id(self.session)

    def list(self):
        self.init_client()
        result = []
        res = self.iam_client.list_policies()
        while True:
            for elem in res["Policies"]:
                arn = elem["Arn"]
                name = policy_arn_to_name(arn, self.account_id)
                if name is None:
                    continue
                result.append(name)
            if not "Marker" in res:
                break
            res = self.iam_client.list_policies(Marker = res["Marker"])
        return result

    def child_handler(self, name):
        self.init_client()
        arn = policy_name_to_arn(name, self.account_id)
        return common_action.NamespaceHandler(
            "conf", ["conf", "policy", "tag"], {
                "conf": PolicyConfHandler(self.iam_client, arn),
                "policy": PolicyDocumentHandler(self.iam_client, arn),
                "tags": PolicyTagsHandler(self.iam_client, arn),
            },
        )

def policy_arn_to_name(arn, account_id):
    m = re.compile("\Aarn:aws:iam::([^:]*):policy/([^:]+)\Z").search(arn)
    if not m:
        return None
    if m.group(1) == account_id:
        return sic_lib.encode_key(m.group(2))
    elif m.group(1) == "aws":
        return sic_lib.encode_key("aws//" + m.group(2))
    else:
        return None

def policy_name_to_arn(name, account_id):
    name = sic_lib.decode_key(name)
    if name.startswith("aws//"):
        return f"arn:aws:iam::aws:policy/{name[5:]}"
    else:
        return f"arn:aws:iam::{account_id}:policy/{name}"

class PolicyConfHandler(common_action.ResourceHandler):
    properties = [
        "Path",
        "PolicyId",
        "Description",
        "IsAttachable",
    ]

    def __init__(self, iam_client, arn):
        self.iam_client = iam_client
        self.arn = arn

    def describe(self):
        try:
            res = self.iam_client.get_policy(PolicyArn=self.arn)
            curr_data = sic_lib.pickup(res["Policy"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

class PolicyDocumentHandler(common_action.ResourceHandler):
    def __init__(self, iam_client, arn):
        self.iam_client = iam_client
        self.arn = arn

    def describe(self):
        try:
            res = self.iam_client.get_policy(PolicyArn=self.arn)
            version_id = res["Policy"]["DefaultVersionId"]
            res = self.iam_client.get_policy_version(PolicyArn=self.arn, VersionId=version_id)
            return res["PolicyVersion"]["Document"]
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise

class PolicyTagsHandler(common_action.ResourceHandler):
    def __init__(self, iam_client, arn):
        self.iam_client = iam_client
        self.arn = arn

    def describe(self):
        try:
            res = self.iam_client.get_policy(PolicyArn=self.arn)
            return sic_lib.decode_tags(res["Policy"]["Tags"])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                return None
            else:
                raise




import os
import random
import sys
import tempfile
import urllib

def removeKey(info: dict, key: str):
    if key in info:
        del info[key]

def exec_diff(content1, content2, dst_file, name1 = "1", name2 = "2"):
    sys.stdout.flush()
    sys.stderr.flush()
    tmpdir = tempfile.mkdtemp()
    fifo1 = tmpdir + "/" + name1
    fifo2 = tmpdir + "/" + name2
    os.mkfifo(fifo1)
    os.mkfifo(fifo2)
    pid = os.fork()
    if pid == 0:
        if dst_file != None:
            sys.stdout = open(dst_file, "w")
        os.execvp("diff", ["diff", "-u", fifo1, fifo2])
    pid1 = os.fork()
    if pid1 == 0:
        writer1 = open(fifo1, "w")
        writer1.write(content1)
        writer1.close()
        sys.exit()
    pid2 = os.fork()
    if pid2 == 0:
        writer2 = open(fifo2, "w")
        writer2.write(content2)
        writer2.close()
        sys.exit()
    os.waitpid(pid1, 0)
    os.waitpid(pid2, 0)
    os.waitpid(pid, 0)
    os.remove(fifo1)
    os.remove(fifo2)
    os.rmdir(tmpdir)

def normalize_script_source(script_source):
    lines = []
    for line in script_source.split("\n"):
        lines.append(line.rstrip(" \t\r"))
    while len(lines) > 0 and lines[0] == "":
        lines = lines[1:]
    while len(lines) > 0 and lines[len(lines) - 1] == "":
        lines = lines[0 : len(lines) - 1]
    return "\n".join(lines) + "\n"

def script_sources_to_yaml(script_dir):
    result = {}
    for file_name in os.listdir(script_dir):
        if os.path.isdir(file_name):
            result[file_name] = script_sources_to_yaml(script_dir + "/" + file_name)
        else:
            with open(script_dir + "/" + file_name) as fh:
                result[file_name] = normalize_script_source(fh.read())
    return result

def script_sources_from_yaml(script_dir, sources):
    for name, data in sources.items():
        path = script_dir + "/" + name
        if isinstance(data, dict):
            if not os.path.exists(path):
                os.makedirs(path)
            script_sources_from_yaml(path, data)
        else:
            with open(path, "w") as fh:
                fh.write(data)

def random_string(length):
    chars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    return ''.join([random.choice(chars) for i in range(length)])

def pickup(src_data, keys):
    if src_data is None:
        return None
    src_data2 = {}
    for key in keys:
        if dictlib_in(src_data, key):
            dictlib_put(src_data2, key, dictlib_get(src_data, key))
    return src_data2

def pickupAndCompareForUpdate(src_data, curr_data, keys):
    src_data2 = {}
    curr_data2 = {}
    for key in keys:
        if dictlib_in(src_data, key):
            dictlib_put(src_data2, key, dictlib_get(src_data, key))
        if dictlib_in(curr_data, key):
            dictlib_put(curr_data2, key, dictlib_get(curr_data, key))
    if src_data2 == curr_data2:
        return None
    return src_data2

def dictlib_in(data, key):
    idx = key.find(".")
    if idx < 0:
        if key in data:
            return True
        else:
            return False
    key2 = key[0:idx]
    if not key2 in data:
        return False
    return dictlib_in(data[key2], key[idx+1:])

def dictlib_get(data, key):
    idx = key.find(".")
    if idx < 0:
        if key in data:
            return data[key]
        else:
            return None
    key2 = key[0:idx]
    if not key2 in data:
        return None
    return dictlib_get(data[key2], key[idx+1:])

def dictlib_put(data, key, value):
    idx = key.find(".")
    if idx < 0:
        data[key] = value
        return
    key2 = key[0:idx]
    if not key2 in data:
        data[key2] = {}
    dictlib_put(data[key2], key[idx+1:], value)

def dict_key_to_lower(info):
    ret = {}
    for key, value in info.items():
        key2 = key[0:1].lower() + key[1:]
        ret[key2] = value
    return ret

def dict_key_capitalize(info):
    ret = {}
    for key, value in info.items():
        key2 = key[0:1].title() + key[1:]
        ret[key2] = value
    return ret

def encode_key(key):
    s = key
    s = urllib.parse.quote(s)
    s = s.replace(".", "%2E")
    s = s.replace("/", "%2F")
    s = s.replace("~", "%7E")
    return s

def decode_key(key):
    s = key
    s = s.replace("%2E", ".")
    s = s.replace("%2F", "/")
    s = s.replace("%7E", "~")
    s = urllib.parse.unquote(s)
    return s

def encode_tags(tags):
    ret = []
    for key, value in tags.items():
        ret.append({"Key": key, "Value": value})
    return ret

def decode_tags(tags):
    ret = {}
    for elem in tags:
        ret[elem["Key"]] = elem["Value"]
    return ret




import re
from setuptools import setup, find_packages

with open('requirements.txt') as requirements_file:
    install_requirements = requirements_file.read().splitlines()

setup(
    name        = "skwadon",
    version     = "0.0.2",
    description = "skwadon",
    author      = "suzuki-navi",
    packages    = find_packages(),
    install_requires = install_requirements,
    include_package_data = True,
    entry_points = {
        "console_scripts": [
            "skwadon = skwadon.main:main",
        ]
    },
    url          = "https://github.com/suzuki-navi/skwadon",
)




import copy
import json

import botocore.exceptions

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action
import skwadon.aws as sic_aws

class JobListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.glue_client = None

    def init_client(self):
        if self.glue_client == None:
            self.glue_client = self.session.client("glue")

    def list(self):
        self.init_client()
        result = []
        res = self.glue_client.get_jobs()
        while True:
            for elem in res['Jobs']:
                name = elem["Name"]
                result.append(name)
            if not "NextToken" in res:
                break
            res = self.glue_client.get_jobs(NextToken = res["NextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        conf_handler = JobConfHandler(self.glue_client, name)
        return common_action.NamespaceHandler(
            "conf", ["conf", "source"], {
            "conf": conf_handler,
            "source": JobSourceHandler(self.session, conf_handler),
            "bookmark": JobBookmarkHandler(self.glue_client, name),
        })

class JobConfHandler(common_action.ResourceHandler):

    properties = [
        "Description",
        "Role",
        "ExecutionProperty",
        "Command",
        "DefaultArguments",
        "NonOverridableArguments",
        "Connections",
        "MaxRetries",
        "Timeout",
        "AllocatedCapacity",
        "MaxCapacity",
        "WorkerType",
        "NumberOfWorkers",
        "SecurityConfiguration",
        "NotificationProperty",
        "GlueVersion",
    ]

    def __init__(self, glue_client, job_name):
        self.glue_client = glue_client
        self.job_name = job_name

    def describe(self):
        try:
            res = self.glue_client.get_job(JobName = self.job_name)
            curr_data = sic_lib.pickup(res["Job"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = sic_lib.pickup(src_data, self.properties)
        update_data["Name"] = self.job_name
        update_data = self._modify_data_for_put(update_data)
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_job(Name = {self.job_name}, ...)",
            lambda:
                self.glue_client.create_job(**update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = sic_lib.pickupAndCompareForUpdate(src_data, curr_data, self.properties)
        if update_data != None:
            update_data = self._modify_data_for_put(update_data)
            sic_main.exec_put(confirmation_flag,
                f"glue_client.update_job(JobName = {self.job_name}, ...)",
                lambda:
                    self.glue_client.update_job(JobName = self.job_name, JobUpdate = update_data)
            )

    def _modify_data_for_put(self, update_data):
        update_data = copy.copy(update_data)
        if "WorkerType" in update_data and update_data["WorkerType"] == "Standard":
            # MaxCapacity が必須で AllocatedCapacity の指定は不可
            sic_lib.removeKey(update_data, "AllocatedCapacity")
        elif "NumberOfWorkers" in update_data:
            sic_lib.removeKey(update_data, "AllocatedCapacity")
            sic_lib.removeKey(update_data, "MaxCapacity")
        else:
            sic_lib.removeKey(update_data, "AllocatedCapacity")
        return update_data

class JobSourceHandler(common_action.ResourceHandler):

    def __init__(self, session, conf_handler):
        self.session = session
        self.conf_handler = conf_handler

    def describe(self):
        conf = self.conf_handler.describe()
        if conf == None:
            return None
        script_s3_path = conf["Command"]["ScriptLocation"]
        script_source = self._fetch_script_source(script_s3_path)
        return script_source

    def create(self, confirmation_flag, src_data):
        if not confirmation_flag:
            # 新規作成時の --dry-run ではアップロード先が不明のため処理をスキップ
            return
        conf = self.conf_handler.describe()
        script_s3_path = conf["Command"]["ScriptLocation"]
        script_source = src_data
        self._put_script_source(confirmation_flag, script_s3_path, script_source)

    def update(self, confirmation_flag, src_data, curr_data):
        conf = self.conf_handler.describe()
        script_s3_path = conf["Command"]["ScriptLocation"]
        script_source = src_data
        self._put_script_source(confirmation_flag, script_s3_path, script_source)

    def _fetch_script_source(self, script_s3_path):
        script_source = sic_aws.fetch_s3_object(self.session, script_s3_path)
        if script_source == None:
            return ""
        script_source = sic_lib.normalize_script_source(script_source)
        return script_source

    def _put_script_source(self, confirmation_flag, script_s3_path, script_source):
        sic_aws.put_s3_object(confirmation_flag, self.session, script_s3_path, script_source)

class JobBookmarkHandler(common_action.ResourceHandler):

    properties = [
        "Version",
        "Run",
        "Attempt",
        "PreviousRunId",
        "RunId",
        "JobBookmark",
    ]

    def __init__(self, glue_client, job_name):
        self.glue_client = glue_client
        self.job_name = job_name

    def describe(self):
        try:
            res = self.glue_client.get_job_bookmark(JobName = self.job_name)
            curr_data = sic_lib.pickup(res["JobBookmarkEntry"], self.properties)
            if "JobBookmark" in curr_data:
                curr_data["JobBookmark"] = json.loads(curr_data["JobBookmark"])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise
        return curr_data




import copy
import json

import botocore

import skwadon.main as sic_main
import skwadon.lib as sic_lib
import skwadon.common_action as common_action

class DatabaseListHandler(common_action.ListHandler):
    def __init__(self, session):
        self.session = session
        self.glue_client = None

    def init_client(self):
        if self.glue_client == None:
            self.glue_client = self.session.client("glue")

    def list(self):
        self.init_client()
        result = []
        res = self.glue_client.get_databases()
        while True:
            for elem in res['DatabaseList']:
                name = elem["Name"]
                result.append(name)
            if not "NextToken" in res:
                break
            res = self.glue_client.get_databases(NextToken = res["NextToken"])
        return result

    def child_handler(self, name):
        self.init_client()
        return common_action.NamespaceHandler(
            "conf", ["conf"], {
            "conf": DatabaseConfHandler(self.glue_client, name),
            "tables": TableListHandler(self.glue_client, name),
        })

class DatabaseConfHandler(common_action.ResourceHandler):

    properties = [
        "Description",
        "LocationUri",
        "Parameters",
        "CreateTableDefaultPermissions",
        "TargetDatabase",
    ]

    def __init__(self, glue_client, database_name):
        self.glue_client = glue_client
        self.database_name = database_name

    def describe(self):
        try:
            res = self.glue_client.get_database(Name = self.database_name)
            curr_data = sic_lib.pickup(res["Database"], self.properties)
            return curr_data
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "EntityNotFoundException":
                return None
            else:
                raise

    def create(self, confirmation_flag, src_data):
        update_data = copy.copy(src_data)
        update_data["Name"] = self.database_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_database(DatabaseInput = {{Name: {self.database_name}, ...}})",
            lambda:
                self.glue_client.create_database(DatabaseInput = update_data)
        )

    def update(self, confirmation_flag, src_data, curr_data):
        update_data = copy.copy(src_data)
        update_data["Name"] = self.database_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.update_database(Name = {self.database_name}, TableInput = {{Name: {self.database_name}, ...}})",
            lambda:
                self.glue_client.update_database(Name = self.database_name,
                DatabaseInput = update_data)
        )

    def delete(self, confirmation_flag, curr_data):
        sic_main.exec_put(confirmation_flag,
            f"glue_client.delete_database(Name = {self.database_name})",
            lambda:
                self.glue_client.delete_database(Name = self.database_name)
        )
        self.described = False

class TableListHandler(common_action.ListHandler):
    def __init__(self, glue_client, database_name):
        self.glue_client = glue_client
        self.database_name = database_name

    def list(self):
        result = []
        res = self.glue_client.get_tables(DatabaseName = self.database_name)
        while True:
            for elem in res['TableList']:
                name = elem["Name"]
                result.append(name)
            if not "NextToken" in res:
                break
            res = self.glue_client.get_tables(DatabaseName = self.database_name, NextToken = res["NextToken"])
        return result

    def child_handler(self, name):
        info = TableInfo(self.glue_client, self.database_name, name)
        return common_action.NamespaceHandler(
            "conf", ["conf", "columns"], {
            "conf": TableConfHandler(info),
            "columns": TableColumnsHandler(info),
        })

class TableInfo:

    properties = [
        "Description",
        "Owner",
        "Retention",
        "StorageDescriptor",
        "PartitionKeys",
        "ViewOriginalText",
        "ViewExpandedText",
        "TableType",
        "Parameters",
        "TargetTable",
    ]

    def __init__(self, glue_client, database_name, table_name):
        self.glue_client = glue_client
        self.database_name = database_name
        self.table_name = table_name
        self.info = None
        self.described = False

    def describe(self):
        if not self.described:
            try:
                res = self.glue_client.get_table(DatabaseName = self.database_name, Name = self.table_name)
                self.info = sic_lib.pickup(res["Table"], self.properties)
            except botocore.exceptions.ClientError as e:
                if e.response["Error"]["Code"] == "EntityNotFoundException":
                    self.info = None
                else:
                    raise
            self.described = True
        return self.info

    def create(self, confirmation_flag, src_data):
        update_data = copy.copy(src_data)
        update_data["Name"] = self.table_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.create_table(DatabaseName = {self.database_name}, TableInput = {{Name: {self.table_name}, ...}})",
            lambda:
                self.glue_client.create_table(DatabaseName = self.database_name,
                TableInput = update_data)
        )
        self.described = False

    def update(self, confirmation_flag, src_data):
        if src_data == self.describe():
            return
        update_data = copy.copy(src_data)
        update_data["Name"] = self.table_name
        sic_main.exec_put(confirmation_flag,
            f"glue_client.update_table(DatabaseName = {self.database_name}, TableInput = {{Name: {self.table_name}, ...}})",
            lambda:
                self.glue_client.update_table(DatabaseName = self.database_name,
                TableInput = update_data)
        )
        self.described = False

    def delete(self, confirmation_flag):
        sic_main.exec_put(confirmation_flag,
            f"glue_client.delete_table(DatabaseName = {self.database_name}, Name = {self.table_name})",
            lambda:
                self.glue_client.delete_table(
                    DatabaseName = self.database_name,
                    Name = self.table_name)
        )
        self.described = False

class TableConfHandler(common_action.ResourceHandler):

    def __init__(self, info: TableInfo):
        self.info = info

    def describe(self):
        info = self.info.describe()
        if info == None:
            return None
        curr_data = copy.copy(info)
        curr_data["StorageDescriptor"] = copy.copy(curr_data["StorageDescriptor"])
        sic_lib.removeKey(curr_data["StorageDescriptor"], "Columns")
        return curr_data

    def create(self, confirmation_flag, src_data):
        update_data = copy.copy(src_data)
        update_data["StorageDescriptor"]["Columns"] = []
        self.info.create(confirmation_flag, update_data)

    def update(self, confirmation_flag, src_data, curr_data):
        info = self.info.describe()
        update_data = copy.copy(src_data)
        update_data["StorageDescriptor"] = copy.copy(src_data["StorageDescriptor"])
        update_data["StorageDescriptor"]["Columns"] = info["StorageDescriptor"]["Columns"]
        self.info.update(confirmation_flag, update_data)

    def delete(self, confirmation_flag, curr_data):
        self.info.delete(confirmation_flag)

class TableColumnsHandler(common_action.ResourceHandler):

    def __init__(self, info: TableInfo):
        self.info = info

    def describe(self):
        info = self.info.describe()
        if info == None:
            return None
        curr_data = self._decode_columns(info["StorageDescriptor"]["Columns"])
        return curr_data

    def create(self, confirmation_flag, src_data):
        if not confirmation_flag:
            # 新規作成時の --dry-run ではdescribeでエラーになるため処理をスキップ
            return
        self.update(confirmation_flag, src_data, None)

    def update(self, confirmation_flag, src_data, curr_data):
        info = self.info.describe()
        info = copy.copy(info)
        info["StorageDescriptor"] = copy.copy(info["StorageDescriptor"])
        info["StorageDescriptor"]["Columns"] = self._encode_columns(src_data)
        self.info.update(confirmation_flag, info)

    def delete(self, confirmation_flag, curr_data):
        pass

    def _decode_columns(self, info):
        data = {}
        for elem in info:
            name = elem["Name"]
            if len(elem) == 2 and "Type" in elem:
                value = elem["Type"]
            else:
                value = copy.copy(elem)
                del value["Name"]
            data[name] = value
        return data

    def _encode_columns(self, data):
        info = []
        for name, value in data.items():
            if isinstance(value, str):
                elem = {
                    "Name": name,
                    "Type": value,
                }
            else:
                elem = copy.copy(value)
                elem["Name"] = name
            info.append(elem)
        return info
